# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest
import json
import msgpack

from flask import url_for

from swh.core.api import remote_api_endpoint, RPCServerApp


@pytest.fixture
def app():
    class TestStorage:
        @remote_api_endpoint('test_endpoint_url')
        def test_endpoint(self, test_data, db=None, cur=None):
            assert test_data == 'spam'
            return 'egg'

        @remote_api_endpoint('path/to/endpoint')
        def something(self, data, db=None, cur=None):
            return data

    return RPCServerApp('testapp', backend_class=TestStorage)


def test_api_endpoint(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Content-Type', 'application/json'),
                 ('Accept', 'application/json')],
        data=json.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/json'


def test_api_nego_default(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Content-Type', 'application/json')],
        data=json.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/json'
    assert res.data == b'"toto"'


def test_api_nego_accept(flask_app_client):
    res = flask_app_client.post(
        url_for('something'),
        headers=[('Accept', 'application/x-msgpack'),
                 ('Content-Type', 'application/x-msgpack')],
        data=msgpack.dumps({'data': 'toto'}),
    )
    assert res.status_code == 200
    assert res.mimetype == 'application/x-msgpack'
    assert res.data == b'\xa4toto'


def test_rpc_server(flask_app_client):
    res = flask_app_client.post(
        url_for('test_endpoint'),
        headers=[('Content-Type', 'application/x-msgpack'),
                 ('Accept', 'application/x-msgpack')],
        data=b'\x81\xa9test_data\xa4spam')

    assert res.status_code == 200
    assert res.mimetype == 'application/x-msgpack'
    assert res.data == b'\xa3egg'



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.core.api import remote_api_endpoint, RPCServerApp, RPCClient
from swh.core.api import error_handler, encode_data_server


# this class is used on the server part
class RPCTest:
    @remote_api_endpoint('endpoint_url')
    def endpoint(self, test_data, db=None, cur=None):
        assert test_data == 'spam'
        return 'egg'

    @remote_api_endpoint('path/to/endpoint')
    def something(self, data, db=None, cur=None):
        return data


# this class is used on the client part. We cannot inherit from RPCTest
# because the automagic metaclass based code that generates the RPCClient
# proxy class from this does not handle inheritance properly.
# We do add an endpoint on the client side that has no implementation
# server-side to test this very situation (in should generate a 404)
class RPCTest2:
    @remote_api_endpoint('endpoint_url')
    def endpoint(self, test_data, db=None, cur=None):
        assert test_data == 'spam'
        return 'egg'

    @remote_api_endpoint('path/to/endpoint')
    def something(self, data, db=None, cur=None):
        return data

    @remote_api_endpoint('not_on_server')
    def not_on_server(self, db=None, cur=None):
        return 'ok'


class RPCTestClient(RPCClient):
    backend_class = RPCTest2


@pytest.fixture
def app():
    # This fixture is used by the 'swh_rpc_adapter' fixture
    # which is defined in swh/core/pytest_plugin.py
    application = RPCServerApp('testapp', backend_class=RPCTest)
    @application.errorhandler(Exception)
    def my_error_handler(exception):
        return error_handler(exception, encode_data_server)
    return application


@pytest.fixture
def swh_rpc_client_class():
    # This fixture is used by the 'swh_rpc_client' fixture
    # which is defined in swh/core/pytest_plugin.py
    return RPCTestClient


def test_api_client_endpoint_missing(swh_rpc_client):
    with pytest.raises(AttributeError):
        swh_rpc_client.missing(data='whatever')


def test_api_server_endpoint_missing(swh_rpc_client):
    # A 'missing' endpoint (server-side) should raise an exception
    # due to a 404, since at the end, we do a GET/POST an inexistent URL
    with pytest.raises(Exception, match='404 Not Found'):
        swh_rpc_client.not_on_server()


def test_api_endpoint_kwargs(swh_rpc_client):
    res = swh_rpc_client.something(data='whatever')
    assert res == 'whatever'
    res = swh_rpc_client.endpoint(test_data='spam')
    assert res == 'egg'


def test_api_endpoint_args(swh_rpc_client):
    res = swh_rpc_client.something('whatever')
    assert res == 'whatever'
    res = swh_rpc_client.endpoint('spam')
    assert res == 'egg'



#

import logging
import textwrap

import click
from click.testing import CliRunner
import pytest

from swh.core.cli import swh as swhmain


help_msg = '''Usage: swh [OPTIONS] COMMAND [ARGS]...

  Command line interface for Software Heritage.

Options:
  -l, --log-level [NOTSET|DEBUG|INFO|WARNING|ERROR|CRITICAL]
                                  Log level (defaults to INFO).
  --log-config FILENAME           Python yaml logging configuration file.
  -h, --help                      Show this message and exit.

Notes:
  If both options are present, --log-level will override the root logger
  configuration set in --log-config.

  The --log-config YAML must conform to the logging.config.dictConfig schema
  documented at https://docs.python.org/3/library/logging.config.html.
'''


def test_swh_help():
    runner = CliRunner()
    result = runner.invoke(swhmain, ['-h'])
    assert result.exit_code == 0
    assert result.output.startswith(help_msg)

    result = runner.invoke(swhmain, ['--help'])
    assert result.exit_code == 0
    assert result.output.startswith(help_msg)


def test_command():
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['test'])
    assert result.exit_code == 0
    assert result.output.strip() == 'Hello SWH!'


def test_loglevel_default(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 20
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['test'])
    assert result.exit_code == 0
    print(result.output)
    assert result.output.strip() == '''Hello SWH!'''


def test_loglevel_error(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 40
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['-l', 'ERROR', 'test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''


def test_loglevel_debug(caplog):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        assert logging.root.level == 10
        click.echo('Hello SWH!')

    runner = CliRunner()
    result = runner.invoke(swhmain, ['-l', 'DEBUG', 'test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''


@pytest.fixture
def log_config_path(tmp_path):
    log_config = textwrap.dedent('''\
    ---
    version: 1
    formatters:
      formatter:
        format: 'custom format:%(name)s:%(levelname)s:%(message)s'
    handlers:
      console:
        class: logging.StreamHandler
        stream: ext://sys.stdout
        formatter: formatter
        level: DEBUG
    root:
      level: DEBUG
      handlers:
        - console
    loggers:
      dontshowdebug:
        level: INFO
    ''')

    (tmp_path / 'log_config.yml').write_text(log_config)

    yield str(tmp_path / 'log_config.yml')


def test_log_config(caplog, log_config_path):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        logging.debug('Root log debug')
        logging.info('Root log info')
        logging.getLogger('dontshowdebug').debug('Not shown')
        logging.getLogger('dontshowdebug').info('Shown')

    runner = CliRunner()
    result = runner.invoke(
        swhmain, [
            '--log-config', log_config_path,
            'test',
        ],
    )

    assert result.exit_code == 0
    assert result.output.strip() == '\n'.join([
        'custom format:root:DEBUG:Root log debug',
        'custom format:root:INFO:Root log info',
        'custom format:dontshowdebug:INFO:Shown',
    ])


def test_log_config_log_level_interaction(caplog, log_config_path):
    @swhmain.command(name='test')
    @click.pass_context
    def swhtest(ctx):
        logging.debug('Root log debug')
        logging.info('Root log info')
        logging.getLogger('dontshowdebug').debug('Not shown')
        logging.getLogger('dontshowdebug').info('Shown')

    runner = CliRunner()
    result = runner.invoke(
        swhmain, [
            '--log-config', log_config_path,
            '--log-level', 'INFO',
            'test',
        ],
    )

    assert result.exit_code == 0
    assert result.output.strip() == '\n'.join([
        'custom format:root:INFO:Root log info',
        'custom format:dontshowdebug:INFO:Shown',
    ])


def test_aliased_command():
    @swhmain.command(name='canonical-test')
    @click.pass_context
    def swhtest(ctx):
        'A test command.'
        click.echo('Hello SWH!')
    swhmain.add_alias(swhtest, 'othername')

    runner = CliRunner()

    # check we have only 'canonical-test' listed in the usage help msg
    result = runner.invoke(swhmain, ['-h'])
    assert result.exit_code == 0
    assert 'canonical-test  A test command.' in result.output
    assert 'othername' not in result.output

    # check we can execute the cmd with 'canonical-test'
    result = runner.invoke(swhmain, ['canonical-test'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''

    # check we can also execute the cmd with the alias 'othername'
    result = runner.invoke(swhmain, ['othername'])
    assert result.exit_code == 0
    assert result.output.strip() == '''Hello SWH!'''



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import logging.config

import click
import pkg_resources
import yaml

LOG_LEVEL_NAMES = ['NOTSET', 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL']

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

logger = logging.getLogger(__name__)


class AliasedGroup(click.Group):
    '''A simple Group that supports command aliases, as well as notes related to
    options'''

    def __init__(self, name=None, commands=None, **attrs):
        self.option_notes = attrs.pop('option_notes', None)
        super().__init__(name, commands, **attrs)

    @property
    def aliases(self):
        if not hasattr(self, '_aliases'):
            self._aliases = {}
        return self._aliases

    def get_command(self, ctx, cmd_name):
        return super().get_command(ctx, self.aliases.get(cmd_name, cmd_name))

    def add_alias(self, name, alias):
        if not isinstance(name, str):
            name = name.name
        self.aliases[alias] = name

    def format_options(self, ctx, formatter):
        click.Command.format_options(self, ctx, formatter)
        if self.option_notes:
            with formatter.section('Notes'):
                formatter.write_text(self.option_notes)
        self.format_commands(ctx, formatter)


@click.group(
    context_settings=CONTEXT_SETTINGS, cls=AliasedGroup,
    option_notes='''\
If both options are present, --log-level will override the root logger
configuration set in --log-config.

The --log-config YAML must conform to the logging.config.dictConfig schema
documented at https://docs.python.org/3/library/logging.config.html.
'''
)
@click.option('--log-level', '-l', default=None,
              type=click.Choice(LOG_LEVEL_NAMES),
              help="Log level (defaults to INFO).")
@click.option('--log-config', default=None,
              type=click.File('r'),
              help="Python yaml logging configuration file.")
@click.pass_context
def swh(ctx, log_level, log_config):
    """Command line interface for Software Heritage.
    """
    if log_level is None and log_config is None:
        log_level = 'INFO'

    if log_config:
        logging.config.dictConfig(yaml.safe_load(log_config.read()))

    if log_level:
        log_level = logging.getLevelName(log_level)
        logging.root.setLevel(log_level)

    ctx.ensure_object(dict)
    ctx.obj['log_level'] = log_level


def main():
    # Even though swh() sets up logging, we need an earlier basic logging setup
    # for the next few logging statements
    logging.basicConfig()
    # load plugins that define cli sub commands
    for entry_point in pkg_resources.iter_entry_points('swh.cli.subcommands'):
        try:
            cmd = entry_point.load()
            swh.add_command(cmd, name=entry_point.name)
        except Exception as e:
            logger.warning('Could not load subcommand %s: %s',
                           entry_point.name, str(e))

    return swh(auto_envvar_prefix='SWH')


if __name__ == '__main__':
    main()
